import { jsx as _jsx } from "react/jsx-runtime";
import { Box, useColorScheme, useTheme } from "@mui/joy";
import CodeMirror from "@uiw/react-codemirror";
import { markdown } from "@codemirror/lang-markdown";
import { languages } from "@codemirror/language-data";
import { syntaxTree } from "@codemirror/language";
import { TextDirectiveParser } from "./directiveParser";
import { GFM } from "@lezer/markdown";
import { RangeSetBuilder, StateField, } from "@codemirror/state";
import { EditorView, Decoration } from "@codemirror/view";
import { BulletListWidget, MarkdownTextDirectiveWidget, MarkdownWidget, TaskWidget, } from "./codeMirrorWidgets";
import { nanoid } from "@reduxjs/toolkit";
import { toolbar } from "./editorToolbar";
const nodeTypeToIgnore = [
    "Document",
    "TableHeader",
    "TableRow",
    "TableCell",
    "TableDelimiter",
    "Paragraph",
    "LinkMark",
    "CodeMark",
    "CodeText",
    "CodeInfo",
    "LinkTitle",
    "LinkLabel",
    "Task",
    "OrderedList",
    "BulletList",
];
/**
 *
 * @param {NerdMarkdownEditorProps} props Component props
 * @returns {ReactElement} NerdMarkdownEditor
 */
export function NerdMarkdownEditor(props) {
    const theme = useTheme();
    const { mode } = useColorScheme();
    return (_jsx(Box, { children: _jsx(Box, { sx: {
                "& table": {
                    border: "1px solid grey",
                    borderRadius: "2px",
                },
                "& td": {
                    border: "1px solid lightgrey",
                    borderRadius: "2px",
                },
                "& p,span,div": {
                    fontStyle: theme.typography["body-md"],
                    fontFamily: theme.typography["body-md"].fontFamily,
                },
                "& h1": theme.typography.h1,
                "& h2": theme.typography.h2,
                "& h3": theme.typography.h3,
                "& h4": theme.typography.h4,
                "& h5": theme.typography["title-lg"],
                "& h6": theme.typography["title-md"],
                "& .ATXHeading1": theme.typography.h1,
                "& .ATXHeading2": theme.typography.h2,
                "& .ATXHeading3": theme.typography.h3,
                "& .ATXHeading4": theme.typography.h4,
                "& .ATXHeading5": theme.typography["title-lg"],
                "& .ATXHeading6": theme.typography["title-md"],
                "& .hiddenMark": {
                    display: "none",
                },
                "& .blockquote": {
                    ml: 4,
                },
                "& .strong": {
                    fontWeight: theme.fontWeight.lg,
                },
                "& .cm-editor": {
                    outline: "none",
                    height: "92vh",
                },
                "& #editor": {
                    boxShadow: (theme) => theme.shadow.sm,
                },
                "& .cm-scroller": { overflow: "auto" },
            }, children: _jsx(CodeMirror, { id: "editor", value: props.value, onChange: props.onChange, height: "80%", basicSetup: {
                    lineNumbers: false,
                    foldGutter: false,
                    highlightActiveLine: false,
                    syntaxHighlighting: false,
                    autocompletion: false,
                    highlightSelectionMatches: false,
                }, theme: mode === "dark" ? "dark" : "light", extensions: [
                    markdown({
                        codeLanguages: languages,
                        extensions: [TextDirectiveParser, GFM],
                    }),
                    myPlugin(props.appTheme),
                    toolbar(),
                ] }) }) }));
}
function getTextWidth(text, font) {
    // re-use canvas object for better performance
    const canvas = document.createElement("canvas");
    const context = canvas.getContext("2d");
    if (context)
        context.font = font;
    const metrics = context?.measureText(text);
    return Math.trunc(metrics?.width || 0);
}
function getCssStyle(element, prop) {
    return window.getComputedStyle(element, null).getPropertyValue(prop);
}
function getCanvasFont(el = document.body) {
    const fontWeight = getCssStyle(el, "font-weight");
    const fontSize = getCssStyle(el, "font-size");
    const fontFamily = getCssStyle(el, "font-family");
    return `${fontWeight} ${fontSize} ${fontFamily}`;
}
const myPlugin = (theme) => StateField.define({
    create(state) {
        return job(state, theme);
    },
    update(_oldState, transaction) {
        return job(transaction.state, theme);
    },
    provide(field) {
        return EditorView.decorations.from(field);
    },
});
/**
 *
 * @param {EditorState} state Editor state
 * @param {Theme} theme App theme
 * @returns {RangeSet<Decoration>} List of decorations
 */
function job(state, theme) {
    const builder = new RangeSetBuilder();
    syntaxTree(state).iterate({
        enter(node) {
            if (!nodeTypeToIgnore.includes(node.type.name)) {
                const isMultiline = state.doc.slice(node.from, node.to).lines > 1;
                const text = state.doc.sliceString(node.from, node.to);
                const cursorLine = state.doc.lineAt(state.selection.main.head);
                if (node.type.name == "TextDirective") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    builder.add(node.from, node.to, Decoration.replace({
                        widget: new MarkdownTextDirectiveWidget(nanoid(), text, theme),
                        inclusive: true,
                    }));
                }
                else if (node.type.name.includes("ATXHeading")) {
                    builder.add(node.from, node.to, Decoration.mark({
                        attributes: {
                            class: node.type.name,
                        },
                    }));
                }
                else if (node.type.name == "ListItem") {
                    let level = 1;
                    let _node = node.node;
                    let parent = node.node;
                    while (_node.type.name !== "Document") {
                        const _parent = _node.parent;
                        if (_parent) {
                            if (_parent?.node.type.name == "ListItem") {
                                parent = _parent.node;
                                level += 1;
                            }
                            _node = _parent?.node;
                        }
                        else
                            break;
                    }
                    const firstChild = node.node.firstChild;
                    const firstChildText = firstChild
                        ? state.doc.sliceString(firstChild.node.from, firstChild.node.to)
                        : "";
                    let offset = 0;
                    if (level > 1) {
                        const text = state.doc.sliceString(parent.firstChild?.node.from || 0, parent.firstChild?.node.to);
                        const font = getCanvasFont(document.getElementById("editor") || undefined);
                        offset = getTextWidth(["*", "-", "+"].includes(text)
                            ? level % 2 == 0
                                ? "◦"
                                : "•"
                            : text, font);
                    }
                    if (firstChild &&
                        ["*", "-", "+"].includes(firstChildText) &&
                        !((isMultiline &&
                            cursorLine.from >= node.from &&
                            cursorLine.from <= node.to) ||
                            (node.from >= cursorLine.from && node.to <= cursorLine.to))) {
                        builder.add(firstChild.node.from, firstChild.node.to, Decoration.replace({
                            widget: new BulletListWidget(nanoid(), text, node.node.parent?.node.parent?.node.type.name == "ListItem" &&
                                node.node.prevSibling
                                ? true
                                : false, level, offset),
                            inclusiveEnd: true,
                        }));
                    }
                    else {
                        builder.add(node.from, node.from + 1, Decoration.mark({
                            attributes: {
                                style: node.node.parent?.node.parent?.node.type.name ==
                                    "ListItem" && node.node.prevSibling
                                    ? `margin-left:${16 * level + offset / 2}px;`
                                    : "margin-left:16px;",
                            },
                            inclusive: false,
                        }));
                    }
                }
                else if (node.type.name == "ListMark") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    const line = state.doc.lineAt(node.from);
                    const listItem = state.doc.sliceString(line.from, line.to);
                    const match = listItem.matchAll(/(-|\+|\*) \[(x|X| )\]/gm).next();
                    if (match.value && match.value.length > 0) {
                        builder.add(node.from, node.to, Decoration.mark({
                            class: "hiddenMark",
                        }));
                    }
                }
                else if (node.type.name == "TaskMarker") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    builder.add(node.from, node.to, Decoration.replace({
                        widget: new TaskWidget(nanoid(), text, node.from, node.to),
                        inclusive: true,
                    }));
                }
                else if (node.type.name == "Blockquote") {
                    builder.add(node.from, node.to, Decoration.mark({
                        class: "blockquote",
                    }));
                }
                else if (node.type.name == "QuoteMark") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    builder.add(node.from, node.to, Decoration.mark({
                        class: "hiddenMark",
                    }));
                }
                else if (node.type.name == "HeaderMark") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    builder.add(node.from, node.to + 1, Decoration.mark({
                        class: "hiddenMark",
                    }));
                }
                else if (node.type.name == "Table") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        builder.add(node.from, node.to, Decoration.mark({
                            tagName: "code",
                        }));
                    else
                        builder.add(node.from, node.to, Decoration.replace({
                            widget: new MarkdownWidget(nanoid(), text),
                            inclusive: true,
                        }));
                }
                else if (node.type.name == "EmphasisMark") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    else
                        builder.add(node.from, node.to, Decoration.mark({
                            class: "hiddenMark",
                        }));
                }
                else if (node.type.name == "StrongEmphasis") {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    else
                        builder.add(node.from, node.to, Decoration.mark({
                            class: "strong",
                        }));
                }
                else {
                    if ((isMultiline &&
                        cursorLine.from >= node.from &&
                        cursorLine.from <= node.to) ||
                        (node.from >= cursorLine.from && node.to <= cursorLine.to))
                        return;
                    builder.add(node.from, node.to, Decoration.replace({
                        widget: new MarkdownWidget(nanoid(), text),
                        inclusive: true,
                    }));
                }
            }
        },
    });
    return builder.finish();
}
